/**
 * ICP localization 
 * Skeleton code for teaching 
 * A3M33MKR
 * Czech Technical University 
 * Faculty of Electrical Engineering
 * Intelligent and Mobile Robotics group
 *
 * Authors: Zdeněk Kasl, Karel Košnar kosnar@labe.felk.cvut.cz
 *
 * Licence: MIT (see LICENSE file)
 **/

#include <cmath>
#include <cassert>
#include <cstdlib>
#include <fstream>
#include <iostream>

#include "gui/gui.h"
#include "dataLoader/laserDataLoader.h"

using namespace imr;
using namespace gui;
using namespace laserDataLoader;


double toRadians(const double alpha);

// Function performing the iterative closest point method - you must to implement it.
RobotPosition icp(const Measurement &reference, const Measurement &actual, const RobotPosition &previousPosition);

// Function convert RobotPosition to Point structure (for drawing).
Point robotPosition2point(const RobotPosition &rp);

// Convert the laser scan into vector of points in Cartesian coordinate system using the odometry from the Measurement structure.
void calculateRawPoints(RawPoints &rp, const Measurement &m);

// Convert the laser scan into vector of points in Cartesian coordinate system using given robot position p.
void calculateRawPoints(RawPoints &rp, const Measurement &m, const RobotPosition &p);

// Convert point in polar coordinate system  into Cartesian coordinate system (possibly moved by Robot Position).
void polar2cartesian(Point &p, double alpha, double r, const RobotPosition &p0);

void help(char** argv)
{
    std::cout << "\nUsage of the program " << argv[0] << ":\n"
         << "Parameter [-h or -H] displays this message.\n"
         << "Parameter [-f or -F] specifies path to data."
         << "Parameter [-m or -M] specifies number of measurements taken,\n"
         << "   default number of measurements is 2.\n"
         << std::endl;
}

int main(int argc, char** argv)
{
    int nMeasurements = 2;
    char *dataFile;

    // argument count must be greater than three
    // >>> at least source file must be specified
    if (argc < 3)
    {
        help(argv);
        return EXIT_FAILURE;
    }

    // Parse all console parameters
    for (int i = 0; i < argc; i++)
    {
        if (argv[i][0] == '-')
        {
            switch (argv[i][1])
            {
                //> HELP
                case 'H' :
                case 'h' :
                    help(argv);
                    break;
                
                //> Source file
                case 'F' :
                case 'f' :
                    assert(i + 1 < argc);
                    dataFile = argv[i + 1];
                    break;

                //> Number of Measurements
                case 'M' :
                case 'm' :
                    assert(i + 1 < argc);
                    assert(atoi(argv[i + 1]) > 1);
                    nMeasurements = atoi(argv[i + 1]);
                    break;
                
                default :
                    std::cout << "Parameter \033[1;31m" << argv[i] << "\033[0m is not valid!\n"
                        << "Use parameter -h or -H for help." << std::endl;
                    break;
            }
        }
    }
    // All parameters parsed.

    std::cout << "Number of measuremetns taken: " << nMeasurements << "\n"
        << "Source file: " << dataFile
        << std::endl;

    // Load data.
    LaserDataLoader loader(dataFile, nMeasurements, "FLASER");
    
    // RawPoints (defined in gui/gui.h) is a vector of Points.
    RawPoints initial;
    RawPoints tentative;

    // Measurement (defined in dataLoader/laserDataLoader.h) is a structure
    // holding the range data and odometry position.
    Measurement previous;

    // RobotPosition (defined in dataLoader/laserDataLoader.h).
    RobotPosition previousPosition;

    // Initialize previous measurement.
    previous = loader[0];
    // Initialize reference scan.
    calculateRawPoints(initial, loader[0]);
    // Initialize previous position .
    previousPosition = loader[0].position;
    // Initialize the next scan.
    calculateRawPoints(tentative, loader[1]);
    // Initial robot position.
    Gui gui(initial, tentative, robotPosition2point(loader[0].position));
    
    // Comment the line below in order to let the program continue right away.
    gui.startInteractor();
    
    for (int i = 1; i < nMeasurements; i++)
    {
        RawPoints tentative;
        // Compute the points of non-corrected measurement.
        calculateRawPoints(tentative, loader[i]);
        // Vizualize tentative non-corrected points (red).
        gui.setTentativePoints(tentative);  

        Measurement actual = loader[i];
        // Calculate the transform and transform data.
        RobotPosition correctedPosition = icp(previous, actual, previousPosition);
        // Correct the actual measurement according to the computed ICP-corrected position.
        RawPoints initial;
        calculateRawPoints(initial, actual, correctedPosition);	

	// Clear map (to show only actually given scan).
	// gui.clearMapPoints();
	
        // Add the new measurement to the map.
	gui.setPointsToMap(initial, robotPosition2point(loader[i].position), 
              robotPosition2point(correctedPosition));

	//> comment line below in order to let the program
	//> continue right away
	if (i % 20 == 0)
           gui.startInteractor();
    }
    gui.clearTentativePoints();
    gui.startInteractor();

    return EXIT_SUCCESS;
}


/// >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
//                               implement ICP here
/// >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>


RobotPosition icp(const Measurement &reference, const Measurement& actual, const RobotPosition& previousPosition) 
{
   RobotPosition ret = actual.position;
   ret.x = ret.x + 1;
   ret.y = ret.y + 1; 
   return ret;
}

/// >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
/// >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>

Point robotPosition2point(const RobotPosition &rp) 
{
   return Point(rp.x, rp.y);
}

void calculateRawPoints(RawPoints &rp, const Measurement &m) 
{
   calculateRawPoints(rp, m, m.position);
}

void calculateRawPoints(RawPoints &rp, const Measurement &m, const RobotPosition &pos)
{
   const double laserResolution = 0.5; // deg
   const double laserShift = -90.0;
   for (unsigned int j = 0; j < m.scan.size(); j++)
   {
      if (m.scan[j] > 50.0)
         continue;
      Point p;
      polar2cartesian(p, toRadians(j * laserResolution + laserShift), m.scan[j], pos);
      rp.push_back(p);
   }
}

void polar2cartesian(Point &p, double alpha, double r, const RobotPosition &p0)
{
   p.x = p0.x+r*cos(alpha+p0.phi);
   p.y = p0.y+r*sin(alpha+p0.phi);
}

double toRadians(const double alpha)
{
   return (alpha*M_PI)/180.0;
}
